/**
 * Open SUIT - Simple User Interface Toolkit
 * 
 * Copyright (c) 2009 France Telecom, http://www.francetelecom.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.ow2.opensuit.xmlmap;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;
import java.util.Vector;

import org.ow2.opensuit.xmlmap.annotations.XmlAncestor;
import org.ow2.opensuit.xmlmap.annotations.XmlAttribute;
import org.ow2.opensuit.xmlmap.annotations.XmlChild;
import org.ow2.opensuit.xmlmap.annotations.XmlChildren;
import org.ow2.opensuit.xmlmap.annotations.XmlContent;
import org.ow2.opensuit.xmlmap.annotations.XmlDoc;
import org.ow2.opensuit.xmlmap.annotations.XmlElement;
import org.ow2.opensuit.xmlmap.annotations.XmlEnumeration;
import org.ow2.opensuit.xmlmap.annotations.XmlParent;
import org.ow2.opensuit.xmlmap.annotations.XmlRoot;
import org.ow2.opensuit.xmlmap.annotations.XmlSimpleType;
import org.ow2.opensuit.xmlmap.annotations.XmlSubstitutionGroup;
import org.ow2.opensuit.xmlmap.mapping.FieldMapping;
import org.ow2.opensuit.xmlmap.mapping.IMappingMethod;
import org.ow2.opensuit.xmlmap.mapping.MappingError;


public class BindWithAnnotations extends IMappingMethod 
{
	// =====================================================================================
	// === Private Stuff
	// =====================================================================================
	private static int JVM_FIELD_ENUM_DIRECTION = 0;// unknown
	private static class TestClass
	{
		private Object field1;
		private Object field2;
	}
	/**
	 * Returns whether the JVM enumerates the declared fields in the same order
	 * than they were declared, or in the opposite order.
	 */
	private static boolean fieldEnumFollowsDeclaration()
	{
		if(JVM_FIELD_ENUM_DIRECTION == 0)
		{
			// --- have to init
			Field[] fields = TestClass.class.getDeclaredFields();
			JVM_FIELD_ENUM_DIRECTION = fields[0].getName().equals("field1") ? 1 : -1;
		}
		return JVM_FIELD_ENUM_DIRECTION == 1;
	}
	// =====================================================================================
	// === Type Methods
	// =====================================================================================
	@Override
	public boolean isSimpleType(Class c)
	{
		return c.isAnnotationPresent(XmlSimpleType.class);
	}
	public boolean isEnumeration(Class c)
	{
		return c.isAnnotationPresent(XmlEnumeration.class);
	}
	public boolean isImportable(Class c)
	{
		XmlSubstitutionGroup annot = (XmlSubstitutionGroup)c.getAnnotation(XmlSubstitutionGroup.class);
		if(annot == null)
			return false;
		return annot.importable();
	}
	public boolean isSubstitutionGroup(Class c)
	{
		return c.isAnnotationPresent(XmlSubstitutionGroup.class);
	}
	public boolean isXmlElement(Class c)
	{
		return c.isAnnotationPresent(XmlElement.class);
	}
	// =====================================================================================
	// === Simple Type Methods
	// =====================================================================================
	@Override
	public Class getBaseSimpleType(Class simpleType)
	{
		XmlSimpleType annot = (XmlSimpleType)simpleType.getAnnotation(XmlSimpleType.class);
		if(annot == null)
			return null;
		return annot.base();
	}
	// =====================================================================================
	// === Element Methods
	// =====================================================================================
	protected void populateMappings(List mappedFields, List mappingErrors, Class iClass, Class iExcludeFrom)
	{
		if(iClass.getSuperclass() != null && iClass.getSuperclass() != iExcludeFrom)
			populateMappings(mappedFields, mappingErrors, iClass.getSuperclass(), iExcludeFrom);

		Field[] fields = iClass.getDeclaredFields();
		boolean fieldsInDeclOrder = fieldEnumFollowsDeclaration();
		for(int i=0; i<fields.length; i++)
		{
            try
            {
            	int idx = fieldsInDeclOrder ? i : fields.length - i - 1;
            	FieldMapping mapping = getXmlMapping(fields[idx]);
				if(mapping != null)
					mappedFields.add(mapping);
            }
            catch(MappingError e)
            {
            	mappingErrors.add(e);
            }
		}
	}
	protected static FieldMapping getXmlMapping(Field field) throws MappingError
	{
		Annotation annot = null;
		if((annot=field.getAnnotation(XmlAttribute.class)) != null)
		{
			// --- Field mapped to an XML attribute
			if(field.getType().isArray() && field.getType() != byte[].class)
				throw new MappingError("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': mapping to an xmlattr involves the field type is not an array.");
			
			// TODO: check type is supported as an attribute (simple types and enumerations)
			
			XmlAttribute xannot = (XmlAttribute)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.ATTRIBUTE);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.required() ? 1 : 0);
			ret.setMaxOccurs(1);
			return ret;
		}
		else if((annot=field.getAnnotation(XmlChild.class)) != null)
		{
			// --- Field mapped to an XML child element
			// TODO: check type is supported as an XML child (simple types, enumerations, substitution groups and xml elements)
			XmlChild xannot = (XmlChild)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(xannot.direct() ? FieldMapping.DIRECT_CHILD_ELT : FieldMapping.CHILD_ELT);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.required() ? 1 : 0);
			ret.setMaxOccurs(1);
			return ret;
		}
		else if((annot=field.getAnnotation(XmlChildren.class)) != null)
		{
			// --- Field mapped to an XML child element
			// TODO: check type is supported as a XML children (vector type of substitution groups and xml elements)
			XmlChildren xannot = (XmlChildren)annot;
			FieldMapping ret = new FieldMapping(field);
			ret.setType(xannot.direct() ? FieldMapping.DIRECT_CHILD_ELT : FieldMapping.CHILD_ELT);
			ret.setName("*".equals(xannot.name()) ? field.getName() : xannot.name());
			ret.setMinOccurs(xannot.minOccurs());
			ret.setMaxOccurs(xannot.maxOccurs());
			return ret;
		}
		else if(field.isAnnotationPresent(XmlContent.class))
		{
//			if(field.getType() != String.class)
//				throw new MappingError("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': xmlcontent field has to be of type String.");
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.CONTENT);
			return ret;
		}
		else if(field.isAnnotationPresent(XmlAncestor.class))
		{
			// --- ancestor field must be at least IXmlObject
//			if(!IXmlObject.class.isAssignableFrom(field.getType()))
//				throw new XmlMappingException("XmlMapping error in field '"+field.getDeclaringClass().getName()+"/"+field.getName()+"': xmlancestor field should necessarily be a subclass of IXmlObject.");
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.ANCESTOR);
			return ret;
		}
		else if(field.isAnnotationPresent(XmlParent.class))
		{
			// --- parent field must be at least IXmlObject
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.PARENT);
			return ret;
		}
		else if(field.isAnnotationPresent(XmlRoot.class))
		{
			// --- root field must be at least IXmlObject
			FieldMapping ret = new FieldMapping(field);
			ret.setType(FieldMapping.ROOT);
			return ret;
		}
		else
		{
			return null;
		}
	}
	// =====================================================================================
	// === Enumeration Methods
	// =====================================================================================
	private Hashtable _class2Fields = new Hashtable();
	
	/**
	 * Returns the number of elements in the enumeration.
	 * @return The number of elements.
	 */
	public int getNbOfEnumItems(Class iEnumClass)
	{
		if(iEnumClass.isEnum())
			return iEnumClass.getEnumConstants().length;
		
		return _getFakeEnumItems(iEnumClass).length;
	}
	/**
	 * Returns an enumeration element from its name.
	 * @param iName The enumeration element name.
	 * @return The enumeration element value.
	 */
	public Object getEnumItem(Class iEnumClass, String iName)
	{
		if(iEnumClass.isEnum())
			return Enum.valueOf(iEnumClass, iName);
		
		Field[] elements = _getFakeEnumItems(iEnumClass);
		for(int i=0; i<elements.length; i++)
		{
			if(elements[i].getName().equals(iName))
			{
	            try
                {
	                return elements[i].get(null);
                }
                catch(Exception e)
                {
                	System.out.println("Enum Error:");
	                e.printStackTrace();
	                return null;
                }
			}
		}
		return null;
	}
	/**
	 * Returns an enumeration element from its index.
	 * @param iIndex The enumeration element index.
	 * @return The enumeration element value.
	 */
	public Object getEnumItem(Class iEnumClass, int iIndex)
	{
		if(iEnumClass.isEnum())
			return iEnumClass.getEnumConstants()[iIndex];
		
		Field[] elements = _getFakeEnumItems(iEnumClass);
		if(iIndex >= elements.length)
			return null;
		try
        {
	        return elements[iIndex].get(null);
        }
        catch(Exception e)
        {
        	System.out.println("Enum Error:");
            e.printStackTrace();
            return null;
        }
	}
	/**
	 * Returns an enumeration element name from its index.
	 * @param iIndex The enumeration element index.
	 * @return The enumeration element name.
	 */
	public String getEnumItemName(Class iEnumClass, int iIndex)
	{
		if(iEnumClass.isEnum())
			return ((Enum)iEnumClass.getEnumConstants()[iIndex]).name();
			
		Field[] elements = _getFakeEnumItems(iEnumClass);
		if(iIndex >= elements.length)
			return null;
		return elements[iIndex].getName();
	}
	public String getEnumItemName(Object iEnumValue)
	{
		if(iEnumValue instanceof Enum)
			return ((Enum)iEnumValue).name();

		Field[] elements = _getFakeEnumItems(iEnumValue.getClass());
		for(int i=0; i<elements.length; i++)
		{
			try
            {
	            if(iEnumValue == elements[i].get(null))
	            	return elements[i].getName();
            }
            catch(Exception e)
            {
            }
		}
		return null;
	}

	/*
	 * Computes Enuymeration elements (by introspection).
	 */
	private Field[] _getFakeEnumItems(Class iEnumClass)
	{
		Field[] elements = (Field[])_class2Fields.get(iEnumClass);
		if(elements == null)
		{
			Vector elementFields = new Vector();
			Field[] fields = iEnumClass.getDeclaredFields();
			for(int i=0; i<fields.length; i++)
			{
				// --- field must be static and of same type as this
				if((fields[i].getModifiers() & Modifier.STATIC) != 0 && fields[i].getType().equals(iEnumClass))
					elementFields.addElement(fields[i]);
			}
			elements = new Field[elementFields.size()];
			elementFields.copyInto(elements);
			_class2Fields.put(iEnumClass, elements);
		}
		return elements;
	}
	
	// =====================================================================================
	// === Documentation Methods
	// =====================================================================================
	public String getInstanceName(Locale locale, Object element)
	{
		return getElementTitle(locale, element.getClass());
	}
	public String getElementTitle(Locale iLocale, Class iComponentClass)
	{
		return iComponentClass.getName();
	}
	public String getElementDetails(Locale iLocale, Class iComponentClass)
	{
		XmlDoc annot = (XmlDoc)iComponentClass.getAnnotation(XmlDoc.class);
		if(annot == null)
			return null;
		return annot.value();
	}
	public String getMappingTitle(Locale locale, FieldMapping mapping)
	{
		return mapping.getName();
	}
	public String getMappingDetails(Locale locale, FieldMapping mapping)
	{
		XmlDoc annot = (XmlDoc)mapping.getField().getAnnotation(XmlDoc.class);
		if(annot == null)
			return null;
		return annot.value();
	}
	public String getEnumItemDetails(Locale locale, Class enumClass, String name)
	{
		try
		{
			Field f = enumClass.getDeclaredField(name);
			XmlDoc annot = (XmlDoc)f.getAnnotation(XmlDoc.class);
			if(annot == null)
				return null;
			return annot.value();
		}
		catch (SecurityException e)
		{
			return name;
		}
		catch (NoSuchFieldException e)
		{
			return name;
		}
	}
}
